use warnings;
use strict;
use Test::More tests => 19;
use Test::Exception;
use Test::MockModule;

use POWER::Event;
use POWER::Event::IO;


my $SIZE = 204800;

POWER::Event::IO::LISTEN->new({
    host        => '0.0.0.0',
    port        => 10080,
    cb          => \&listener_server,
    wait_for    => $ACCEPTED,
});
POWER::Event::IO::LISTEN->new({
    host        => '0.0.0.0',
    port        => 13128,
    cb          => \&listener_proxy,
    wait_for    => $ACCEPTED,
});

sub new_client {
    my ($port, $cb) = @_;
    POWER::Event::IO->new({
        host        => '127.0.0.1',
        port        => $port,
        cb          => $cb,
        wait_for    => $EOF,
        out_buf     => ('x' x $SIZE),
        in_buf_limit=> 0,
    });
}
new_client(10080, \&client);

my @Wait = (
    ['listener_server', $ACCEPTED   ],
    ['server',          $IN         ],  # depend on system buffers size :(
    ['server',          $IN         ],
    ['server',          $IN         ],
    ['server',          $IN         ],
    ['client',          $EOF        ],
    
    ['listener_proxy',  $ACCEPTED   ],
    ['listener_server', $ACCEPTED   ],
    ['server',          $IN         ],  # depend on system buffers size :(
    ['server',          $IN         ],
    ['server',          $IN         ],
    ['server',          $IN         ],
    ['client2',         $EOF        ],
);

eval { loop() };
die "out of loop: $@";

sub listener_server {
    my ($io) = @_;
    die "listener_server unexpected event: $io->{err}" if $io->{events} != $ACCEPTED;
    is_deeply(shift @Wait, ['listener_server', $io->{events}], 'listener_server');
    POWER::Event::IO->new({
        fh          => delete($io->{in_buf}),
        cb          => \&server,
        wait_for    => $IN,
    });
}

sub server {
    my ($io) = @_;
    die "server error: $io->{err}" if $io->{events} & $ERR;
    if ($io->{events} & $SENT) {
        $io->close();
        return;
    }
    is_deeply(shift @Wait, ['server', $io->{events}], 'server');
    if ($io->{events} & $IN) {
        $io->write('y' x length($io->{in_buf}));
        $io->{in_buf} = q{};
        if ($io->{in_bytes} == $SIZE) {
            if (length($io->{out_buf})) {
                $io->wait_for($SENT);
            } else {
                $io->close();
            }
        }
    }
}

sub listener_proxy {
    my ($io) = @_;
    die "listener_proxy unexpected event: $io->{err}" if $io->{events} != $ACCEPTED;
    is_deeply(shift @Wait, ['listener_proxy', $io->{events}], 'listener_proxy');
    my $proxy_client = POWER::Event::IO->new({
        fh          => delete($io->{in_buf}),
        cb          => \&proxy_client,
    });
    my $proxy_server = POWER::Event::IO->new({
        host        => '127.0.0.1',
        port        => 10080,
        cb          => \&proxy_server,
        wait_for    => $CONNECTED|$IN|$EOF,
        Client      => $proxy_client,
    });
    $proxy_client->{Server} = $proxy_server;
}

sub proxy_client {
    my ($io) = @_;
    die "proxy_client error: $io->{err}" if $io->{events} & $ERR;
    if ($io->{events} & $SENT) {
        $io->close();
        return;
    }
#    is_deeply(shift @Wait, ['proxy_client', $io->{events}], 'proxy_client');
    if ($io->{events} & $IN) {
        $io->{Server}->write($io->{in_buf});
        $io->{in_buf} = q{};
    }
    if ($io->{events} & $EOF) {
        if (length($io->{Server}->{out_buf})) {
            $io->{Server}->wait_for($IN|$SENT);
        } else {
            shutdown $io->{Server}->{fh}, 1;
        }
    }
}

sub proxy_server {
    my ($io) = @_;
    die "proxy_server error: $io->{err}" if $io->{events} & $ERR;
    if ($io->{events} == $CONNECTED) {
        $io->{Client}->wait_for($IN|$EOF);
        return;
    }
    if ($io->{events} & $SENT) {
        shutdown $io->{fh}, 1;
        return;
    }
#    is_deeply(shift @Wait, ['proxy_server', $io->{events}], 'proxy_server');
    if ($io->{events} & $IN) {
        $io->{in_buf} =~ s/y/z/g;
        $io->{Client}->write($io->{in_buf});
        $io->{in_buf} = q{};
    }
    if ($io->{events} & $EOF) {
        if (length($io->{Client}->{out_buf})) {
            $io->{Client}->wait_for($SENT);
        } else {
            $io->{Client}->close();
        }
        $io->{Client}->{Server} = undef;
        $io->{Client}           = undef;
        $io->close();
    }
}

sub client {
    my ($io) = @_;
    die "client unexpected event: $io->{err}" if $io->{events} != $EOF;
    is_deeply(shift @Wait, ['client', $io->{events}], 'client');
    like($io->{in_buf}, qr/\Ay+\z/xms, '  got /\Ay+\z/ reply');
    is(length($io->{in_buf}), $SIZE, "  in_buf size is $SIZE");
    ok($io->{is_eof},                '  {is_eof} set');
    $io->close();
    new_client(13128, \&client2);
}

sub client2 {
    my ($io) = @_;
    die "client2 unexpected event: $io->{err}" if $io->{events} != $EOF;
    is_deeply(shift @Wait, ['client2', $io->{events}], 'client2');
    like($io->{in_buf}, qr/\Az+\z/xms, '  got /\Ay+\z/ reply');
    is(length($io->{in_buf}), $SIZE, "  in_buf size is $SIZE");
    ok($io->{is_eof},                '  {is_eof} set');
    exit;
}

